///////////////////////////////////////////////////////////////////////////
//      MOTU Midi Time Piece BeOS Driver
//		Executed by: Alex Mouton <tmouton @ falcon.csc.calpoly.edu> 1 May, 2001
//		This code was adapted from the two previous drivers:
//		The ALSA driver produced by Michael T. Mayers Jan 09, 2000
//      mail: tweakoz@pacbell.net
//		The BSD driver produced by John Galbraith
///////////////////////////////////////////////////////////////////////////
//      This program is free software; you can redistribute it and/or modify
//      it under the terms of the GNU General Public License as published by
//      the Free Software Foundation; either version 2 of the License, or
//      (at your option) any later version.
//
//      This program is distributed in the hope that it will be useful,
//      but WITHOUT ANY WARRANTY; without even the implied warranty of
//      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//      GNU General Public License for more details.
//
//      You should have received a copy of the GNU General Public License
//      along with this program; if not, write to the Free Software
//      Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
///////////////////////////////////////////////////////////////////////////
//      This driver is for the 'Mark Of The Unicorn' (MOTU) Midi Time Piece AV series of 
//		multiport MIDI interfaces 
//
//      IOPORTS
//      ///////////
//      8 MIDI Ins and 8 MIDI outs
//      Video Sync In (BNC), Word Sync Out (BNC),  			//UNSUPPORTED
//      ADAT Sync Out (DB9) 								//UNSUPPORTED
//      SMPTE in/out (1/4")
//      2 programmable pedal/footswitch inputs and 4 programmable MIDI controller knobs. 	//UNSUPPORTED
//      Macintosh RS422 serial port							//UNSUPPORTED
//      RS422 "network" port for ganging multiple MTP's 	//UNSUPPORTED
//      PC Parallel Port ( which this driver currently uses )
//
//      MISC FEATURES
//      ///////////////
//      Hardware MIDI routing, merging, and filtering   
//      MIDI Synchronization to Video, ADAT, SMPTE and other Clock sources
//      128 'scene' memories, recallable from MIDI program change
//////////////////////////////////////////////////////////////////////////
//		I have produced this driver with only a Midi Express XT as functional reference.
//		I do not plan to implement any extra functionality of the other devices on my own
//		time, but I will work to integrate any changes you've got.  The structure should
//		be there, to just uncomment a few things, and your additional ports will work.
//
//		This device assumes your parallel port is on base address 0x378, and IRQ7.
//		If your parallel port is something different, change MTPAV_IOBASE and MTPAV_IRQ
//		and recompile.
//
//		At the time of release 0.1, midi in has been left unimplemented. Large chunks of it
//		are there, but I havnt got to integrating them yet.
//
//		your feedback, questions, and comments are greatly appreciated,
//		feel free to email me.
//		-Alex Mouton
///////////////////////////////////////////////////////////////////////////

#include <Errors.h>

#include <be/kernel/OS.h> 
#include <be/drivers/KernelExport.h>
#include <be/drivers/Drivers.h>
#include <be/drivers/module.h>
#include <be/drivers/ISA.h>

#include <be/support/SupportDefs.h>

#include <malloc.h>
#include <string.h>
#include <stdlib.h>
#include <fcntl.h>

#define NODEBUG bool __debug__ = set_dprintf_enabled( false ); 
#define UNDEBUG set_dprintf_enabled( __debug__ );
#define DEBUG bool __debug__ = set_dprintf_enabled( true );


// io resources (make these module options?)
#define MTPAV_IOBASE 0x378
#define MTPAV_IRQ    7

#define DREG 0x000
#define SREG 0x001
#define CREG 0x002


// parallel port usage masks
#define SIGS_BYTE 0x08
#define SIGS_RFD 0x80
#define SIGS_IRQ 0x40
#define SIGS_IN0 0x10
#define SIGS_IN1 0x20

#define SIGC_WRITE 0x04
#define SIGC_READ 0x08
#define SIGC_INTEN 0x10

#define SREG_MASK 0xf8
#define CREG_MASK 0x1c



#define MTPAV_NO_CMD 0x00
#define MTPAV_UNKNOWN_STATUS_CMD 0x01

// known commands
#define MTPAV_SYSEX_CMD 0xf0
#define MTPAV_SYSEX_CMD_END 0xf7
#define MTPAV_PORT_CHANGE_CMD 0xf5
#define MTPAV_ACTIVE_SENSE_CMD 0xfe

//
#define MTPAV_MODE_CLOSED			0x00
#define MTPAV_MODE_INPUT_OPENED		0x01
#define MTPAV_MODE_OUTPUT_OPENED	0x02
#define MTPAV_MODE_INPUT_TRIGGERED	0x04
#define MTPAV_MODE_OUTPUT_TRIGGERED	0x08

// possible hardware ports (selected by 0xf5 port message)
//      0x00		all ports
//      0x01 .. 0x08    this MTP's ports 1..8
//      0x09 .. 0x10    networked MTP's ports (9..16) NOT IMPLEMENTED
//      0x11            networked MTP's computer port NOT IMPLEMENTED
//      0x63            to ADAT NOT IMPLEMENTED
#define NUMPORTS (0x08+1)

//buffers are perport
#define DATA_BUF_SIZE 2048

#define STATUSBYTE( x ) ( x >= 0x80 )

#define TOUCH(x) ( (void)x )
#define PORTOPEN(x) ( ( x->mode & MTPAV_MODE_OUTPUT_OPENED ) || ( x->mode & MTPAV_MODE_INPUT_OPENED ) )
#define PORTREADOPEN(x) ( x->mode & MTPAV_MODE_INPUT_OPENED )

#define DEVICE_NAME_MIDI1 "midi/mtpav/1"
#define DEVICE_NAME_MIDI2 "midi/mtpav/2"
#define DEVICE_NAME_MIDI3 "midi/mtpav/3"
#define DEVICE_NAME_MIDI4 "midi/mtpav/4"
#define DEVICE_NAME_MIDI5 "midi/mtpav/5"
#define DEVICE_NAME_MIDI6 "midi/mtpav/6"
#define DEVICE_NAME_MIDI7 "midi/mtpav/7"
#define DEVICE_NAME_MIDI8 "midi/mtpav/8"

// content of *cookie
typedef struct SmtpPort {
	uint8 hwport;
	uint8 mode;
	
	uint8 databuf[ DATA_BUF_SIZE ];
	int16 databufend;

	uint32 flags;
	
	sem_id read_sem;
	
	int16 firstbyte; // pointer to oldest byte
	int16 lastbyte; // pointer to newest byte  
} mtpPort;

typedef struct Smtp {

	spinlock lock;
	
	sem_id write_sem;
	
	mtpPort* ports[NUMPORTS];

	uint8 inmidiport;
	uint8 inmidistate;
	uint8 outmidihwport;
	
	uint8 open_ref_count;
} mtp;
		 
	
status_t init_hardware( void );
status_t init_driver( );
const char** publish_devices( );

device_hooks* find_device(const char* name);
status_t mtpav_open (const char *name, uint32 flags, void** cookie);
int32 mtpav_interrupt (void *data);
status_t mtpav_read (void* cookie, off_t position, void *buf, size_t* num_bytes);
status_t mtpav_write (void* cookie, off_t position, const void* buffer, size_t* num_bytes);
status_t mtpav_control (void* cookie, uint32 op, void* arg, size_t len);
status_t mtpav_close (void* cookie);
status_t mtpav_free (void* cookie);

size_t move_data( mtpPort* port, void* buffer, size_t num_bytes);

int32 mtpav_read_bytes();

void mtpav_enable_interrupt();
void mtpav_disable_interrupt();

void queue_byte( uint8 inbyte );

void mtpav_portscan(void);

uint8 mtpav_getreg(uint16 reg);

void mtpav_putdreg( uint8 val );
void mtpav_putcreg( uint8 val );

void mtpav_send_byte( uint8 byte );

status_t outb( int , const uint8);
uint8 inb( uint16 );

int translate_hwport_to_portindex(int hwport);
int translate_devicename_to_portindex( const char* name );




//	null-terminated array of device names supported by this driver
static const char *mtpav_name[] = {
	DEVICE_NAME_MIDI1,
	DEVICE_NAME_MIDI2,
	DEVICE_NAME_MIDI3,
	DEVICE_NAME_MIDI4,
	DEVICE_NAME_MIDI5,
	DEVICE_NAME_MIDI6,
	DEVICE_NAME_MIDI7,
	DEVICE_NAME_MIDI8,			
	NULL
};


//  function pointers for the device hooks entry points
device_hooks mtpav_hooks = {
	mtpav_open, 			/* -> open entry point */
	mtpav_close, 			/* -> close entry point */
	mtpav_free,			/* -> free cookie */
	mtpav_control, 		/* -> control entry point */
	mtpav_read,			/* -> read entry point */
	mtpav_write,			/* -> write entry point */
	/* Leave select/deselect/readv/writev undefined. The kernel will
	 * use its own default implementation. The basic hooks above this
	 * line MUST be defined, however. */
		NULL,
		NULL,
		NULL,
		NULL
};

int32 api_version = B_CUR_DRIVER_API_VERSION;

struct isa_module_info *isa;
	
static mtp *mtp_card;

status_t init_hardware( void ) {
	NODEBUG
	//kprintf("init_hardware\n");
	UNDEBUG
	return B_OK;
}

//	init_driver - optional function - called every time the driver
//	is loaded.

status_t init_driver( ) {	
	int ix;
	NODEBUG
	
	kprintf("init_driver\n");

	get_module(B_ISA_MODULE_NAME , &isa);
		
	// actually find hardware, assumed true;
	//if no hardware return ENODEV
	//else if error state return ENOSYS or ENOENT
		
	// malloc driver global data
	if ( NULL == (mtp_card = calloc( sizeof(mtp), 1 )) ) {
		kprintf("mtp_card calloc failed!\n");
		return B_NO_MEMORY;
	}

	mtp_card->lock = 0;  //card initially unlocked
	mtp_card->write_sem = create_sem(1, "motu midi express XT write");

	for(ix = 0; ix < NUMPORTS; ix++) {
	
		mtp_card->ports[ix] = malloc( sizeof(mtpPort) );
		mtp_card->ports[ix]->hwport = ix;
		mtp_card->ports[ix]->mode = MTPAV_MODE_CLOSED;

		mtp_card->ports[ix]->databufend = 0;

		mtp_card->ports[ix]->flags = 0;

		mtp_card->ports[ix]->read_sem = create_sem(0, "mtpport read_sem");
		set_sem_owner( mtp_card->ports[ix]->read_sem,  B_SYSTEM_TEAM );
			
		mtp_card->ports[ix]->firstbyte = 0; 
		mtp_card->ports[ix]->lastbyte = -1;
	  
	}

	mtp_card->inmidiport = -1; 
	mtp_card->inmidistate = MTPAV_NO_CMD; 
	mtp_card->outmidihwport = -1; 

	mtp_card->open_ref_count = 0;

	mtpav_portscan();

	UNDEBUG		
	return B_OK;
}







/* ----------
	publish_devices - return a null-terminated array of devices
	supported by this driver.
----- */

const char**
publish_devices()
{
	NODEBUG
	//kprintf("publish_devices\n");
	
	UNDEBUG
	return mtpav_name;
}





/* ----------
	find_device - return ptr to device hooks structure for a
	given device name
----- */

device_hooks* find_device(const char* name)
{
	uint8 ix;
	NODEBUG
	
	//kprintf("find_device %s\n", name);

	for(ix=0; ix < NUMPORTS ; ix++) {
		if (!strcmp(name, mtpav_name[ix])) {
			UNDEBUG
			return &mtpav_hooks;
		}
	}

	UNDEBUG		
	return NULL;
}









/* ----------
	mtpav_open - handle open() calls
----- */

status_t
mtpav_open (const char *name, uint32 flags, void** cookie)
{
	uint8 portindex;

	NODEBUG
	cpu_status cpuformer; 

	kprintf("mtpav_open ");
		
	cpuformer = disable_interrupts();

	portindex = translate_devicename_to_portindex( name );

	kprintf("%d\n", portindex );
	
	//only allow one simultaneous opening of any one mtp port
	if( PORTOPEN( mtp_card->ports[ portindex ] )  ) {
	    kprintf("port opened twice! failed!\n");
		
    	restore_interrupts( cpuformer );
		
		UNDEBUG
		return B_BUSY;
	}
	
	// let port know that its opened
	if( flags & O_RDONLY ) mtp_card->ports[ portindex ]->mode = MTPAV_MODE_INPUT_OPENED;
	else if( flags & O_WRONLY ) mtp_card->ports[ portindex ]->mode = MTPAV_MODE_OUTPUT_OPENED;
	else if( flags & O_RDWR ) mtp_card->ports[ portindex ]->mode = MTPAV_MODE_INPUT_OPENED | MTPAV_MODE_OUTPUT_OPENED;

	mtp_card->ports[ portindex ]->read_sem = create_sem(0, "MOTU Midi Express XT Read");
	set_sem_owner( mtp_card->ports[ portindex ]->read_sem,  B_SYSTEM_TEAM );
	mtp_card->ports[ portindex ]->flags = flags;
			
	// enable the mtp interrupt if this is the only port open
	if( mtp_card->open_ref_count == 0) {
		mtpav_enable_interrupt();
	}
	mtp_card->open_ref_count++;
	
	*cookie =  mtp_card->ports[ portindex ];	
	TOUCH(flags);

	restore_interrupts( cpuformer );

	UNDEBUG	
	return B_OK;
}




int32 mtpav_interrupt (void *data) {
	NODEBUG
	uint8 sbyt;

	int32 returncode;
		
	// SMP safety
	acquire_spinlock(&mtp_card->lock);

	sbyt = mtpav_getreg( SREG );

	// This interrupt is not shared.
	// Means to distinguish a MOTU dataless interrupt and the interrupt of another device is unknown.
	// the MOTUS produce dataless interrupts quite frequently
	//	if this is a dataless interrupt, handle it but do nothing.

	if (!(sbyt & SIGS_BYTE)){
			//kprintf("*");
			release_spinlock(&mtp_card->lock);		
			UNDEBUG
			return B_HANDLED_INTERRUPT;	
	}
	
	//kprintf(".");

	returncode = mtpav_read_bytes( );
	release_spinlock( &mtp_card->lock );	

	UNDEBUG
	return returncode;
}



/* ----------
	mtpav_read - handle read() calls
----- */

status_t
mtpav_read (void* cookie, off_t position, void *buf, size_t* num_bytes)
{
	NODEBUG
	mtpPort *port = (mtpPort*)cookie;
	cpu_status cpuformer;
	status_t sem_err;
	
	//kprintf( "mtpav_read\n" );
	*num_bytes = 0;				// tell caller nothing was read 
	
	if ( port->mode == MTPAV_MODE_CLOSED ) return B_ERROR;
	
	sem_err = acquire_sem( port->read_sem );
	
	if( sem_err == B_OK ) {

		cpuformer = disable_interrupts(); 
		
		//kprintf("acquired read_sem\n");	
		
		*num_bytes = move_data( port, buf, *num_bytes);
	
		restore_interrupts( cpuformer );
	
	}	
	else {

		kprintf("read_sem error\n");
		UNDEBUG
		return B_INTERRUPTED;
	}

		
	UNDEBUG
	return B_OK;

}


size_t
move_data( mtpPort* port, void* buffer, size_t num_bytes) {
	NODEBUG
	size_t bytes = 0;
	status_t lock_err;
		
	lock_err = lock_memory(buffer, num_bytes , B_READ_DEVICE);
	
	if( lock_err >= B_OK ) {

		bytes = 1;
		memcpy( buffer, port->databuf+port->firstbyte, 1 );
		unlock_memory(buffer, num_bytes, B_READ_DEVICE);
				
	}		
	
	port->firstbyte += bytes;
	if( port->firstbyte >= DATA_BUF_SIZE ) port->firstbyte = port->firstbyte - DATA_BUF_SIZE;

	kprintf("pop %x %x\n", port->firstbyte, port->lastbyte );
	
	UNDEBUG	
	return bytes;
}




/* ----------
	mtpav_write - handle write() calls
----- */

status_t
mtpav_write (void* cookie, off_t position, const void* buffer, size_t* num_bytes)
{
	uint8 outbyte = 0;
	uint8 sentcount = 0;

	status_t sem_err;
	
	NODEBUG
	
	mtpPort *port = (mtpPort*)cookie;
	unsigned long flags;

	//kprintf("mtpav_write\n");

	if ( port->mode == MTPAV_MODE_CLOSED ) {
		*num_bytes = 0;
		return B_ERROR;
	}

	sem_err = acquire_sem( mtp_card->write_sem );
	
	if( sem_err == B_OK ) {
	
		// send port change command if necessary 
	 	if( mtp_card->outmidihwport != port->hwport ) {
			//kprintf("mtpav_write port change\n");
			mtpav_send_byte( MTPAV_PORT_CHANGE_CMD );
			mtpav_send_byte( port->hwport );
			mtp_card->outmidihwport = port->hwport;
		}
	
		///////////////////
		//      send data
	
		sentcount = 0;
		while (sentcount < *num_bytes) {
			outbyte = *( (uint8*) buffer + sentcount );
			mtpav_send_byte( outbyte );
			sentcount++;
			snooze(200);
		}
		
		release_sem( mtp_card->write_sem );
		
	}	

	UNDEBUG		
	return B_OK;
}


/* ----------
	mtpav_control - handle ioctl calls
----- */

status_t
mtpav_control (void* cookie, uint32 op, void* arg, size_t len)
{	
	NODEBUG
	
	TOUCH(len);
	kprintf("mtpav_control\n");

	if (op == 'set ') {
		UNDEBUG
		return B_OK;
	}

	UNDEBUG
		
	return ENOSYS;
}


/* ----------
	mtpav_close - handle close() calls
----- */

status_t
mtpav_close (void* cookie)
{
	NODEBUG
	mtpPort *port = (mtpPort*)cookie;	

	kprintf("mtpav_close\n");

	port->mode = MTPAV_MODE_CLOSED;

	delete_sem( port->read_sem );
	port->read_sem = create_sem(0, "mtpport read_sem");
	set_sem_owner( port->read_sem,  B_SYSTEM_TEAM );
			
	UNDEBUG
	return B_OK;
}



/* -----
	mtpav_free - called after the last device is closed, and after all i/o is complete.
----- */

status_t
mtpav_free (void* cookie)
{
	NODEBUG
	
	kprintf( "mtpav_free\n" );
	
	mtp_card->open_ref_count--;

	if( mtp_card->open_ref_count <= 0) {
		mtp_card->open_ref_count = 0;
		mtpav_disable_interrupt();
	}

	return B_OK;
}






void mtpav_wait_rfdhi(void)
{
	NODEBUG
	uint8 sbyte;
	
	kprintf("mtpav_waitrfdi");
	
	sbyte = mtpav_getreg(SREG);
	while (!(sbyte & SIGS_RFD)) {
		kprintf(".");
		sbyte = mtpav_getreg(SREG);
	}
	
	kprintf("\n");
	UNDEBUG
}




// mtpav control
void 
mtpav_portscan (void)	// put mtp into smart routing mode
{
	NODEBUG
	uint8 port;
	
	//kprintf("mtpav_portscan\n");
	
	for (port = 1; port <= 8; port++) {
		mtpav_send_byte( 0xf5 );		
		mtpav_send_byte( port );
		mtpav_send_byte( 0xff );
		mtpav_send_byte( 0xfe );
	}
	UNDEBUG	
}





//	uninit_driver - optional function - called every time the driver
//	is unloaded
void uninit_driver( ) {
	int ix;
	NODEBUG
	
	kprintf( "uninit_driver\n");
	
	for(ix = 0; ix < NUMPORTS; ix++) {
		delete_sem( mtp_card->ports[ix]->read_sem );
		free( mtp_card->ports[ix]);
	}

	delete_sem( mtp_card->write_sem );
			
	// malloc driver global data
	free( mtp_card ); 
	
	put_module( B_ISA_MODULE_NAME );
	UNDEBUG
}





int32 mtpav_read_bytes( )
{
	NODEBUG
	int32 rc = B_HANDLED_INTERRUPT;
	uint8 clrread, setread;
	uint8 mtp_read_byte;
	uint8 read_sr0, read_sr1, read_sr2, read_sr3;
	uint8 cbyt;

	uint8 sbyt = mtpav_getreg(SREG);

	mtpPort *port = mtp_card->ports[ translate_hwport_to_portindex ( mtp_card->inmidiport ) ];
	bool status;

	cbyt = mtpav_getreg(CREG);
	
	clrread = cbyt & (SIGC_READ ^ 0xff);
	setread = cbyt | SIGC_READ;

		//assemble byte from 2bit nibbles
	do {

		mtpav_putcreg( setread );
		read_sr0 = mtpav_getreg(SREG);
		mtpav_putcreg( clrread );

		mtpav_putcreg( setread );
		read_sr1 = mtpav_getreg(SREG);
		mtpav_putcreg( clrread );

		mtpav_putcreg( setread );
		read_sr2 = mtpav_getreg(SREG);
		mtpav_putcreg( clrread );

		mtpav_putcreg( setread );
		read_sr3 = mtpav_getreg(SREG);
		mtpav_putcreg( clrread );

		read_sr0 = ( read_sr0 & (SIGS_IN0 | SIGS_IN1) ) >> 4;
		read_sr1 = ( read_sr1 & (SIGS_IN0 | SIGS_IN1) ) >> 2;
		read_sr2 = ( read_sr2 & (SIGS_IN0 | SIGS_IN1) );
		read_sr3 = ( read_sr3 & (SIGS_IN0 | SIGS_IN1) ) << 2;

		mtp_read_byte = read_sr0 | read_sr1 | read_sr2 | read_sr3;
	
		// watch the raw bytes come in, dont show any bytes 0xfe. 0xfe is the value of a 
		// heartbeat signal sent by some midi implementations, and, when present, show up very frequently
		//if( mtp_read_byte != 0xfe ) kprintf(" %2.2x", mtp_read_byte );

		status = STATUSBYTE( mtp_read_byte );
				
		if( status && mtp_card->inmidistate != MTPAV_NO_CMD ) {
			mtp_card->inmidistate = MTPAV_NO_CMD;
		}
	
		if( status ) {  //this is the beginning of a new message
	
			if( mtp_read_byte == MTPAV_ACTIVE_SENSE_CMD ) {
					//kprintf(".");
					// DONT PASS THE CMD ON TO MIDI SYSTEM.
			}
			else if ( mtp_read_byte == MTPAV_PORT_CHANGE_CMD ) { // Port Change Status Byte
				//kprintf("PORT_CHANGE_CMD %2.2x", mtp_read_byte);
				mtp_card->inmidistate = MTPAV_PORT_CHANGE_CMD;
			}
			else if ( mtp_read_byte == MTPAV_SYSEX_CMD ) {	 // Sysex Status Byte
				//kprintf("SYSEX_CMD %2.2x", mtp_read_byte);
				mtp_card->inmidistate = MTPAV_SYSEX_CMD;
				queue_byte( mtp_read_byte );
			}
			else if( (( 0xF0 & mtp_read_byte ) == 0xC0) ||  (( 0xF0 & mtp_read_byte ) == 0xD0) ) {  //program change or aftertouch cmd Status Bute
				//kprintf("2 BYTE %2.2x", mtp_read_byte);
				mtp_card->inmidistate = MTPAV_UNKNOWN_STATUS_CMD;
				queue_byte( mtp_read_byte );
			}
			else if( (0x80 & mtp_read_byte ) == 0x80 ) {  // Other Status Byte
				//kprintf("3 BYTE %2.2x", mtp_read_byte);
				mtp_card->inmidistate = MTPAV_UNKNOWN_STATUS_CMD;
				queue_byte( mtp_read_byte ); 
			}	
	
		}
		else {	 // Data byte
	
			if (mtp_card->inmidistate == MTPAV_PORT_CHANGE_CMD) {  // handle PORT_CHANGE_CMD data
				//kprintf(" %2.2x \n", mtp_read_byte);
				mtp_card->inmidiport = mtp_read_byte;
				mtp_card->inmidistate = MTPAV_NO_CMD;
			}	
			else if ( (mtp_card->inmidistate == MTPAV_SYSEX_CMD) && (mtp_read_byte == MTPAV_SYSEX_CMD_END) ) {  // handle SYSEX data
				//kprintf(" %2.2x", mtp_read_byte);
				queue_byte( mtp_read_byte );
				//kprintf(" END\n");
				mtp_card->inmidistate = MTPAV_NO_CMD;
				rc = B_INVOKE_SCHEDULER;	
			}
			else {	// normal data
				//kprintf(" %2.2x", mtp_read_byte);
				queue_byte( mtp_read_byte ); 
				rc = B_INVOKE_SCHEDULER;	
			}
	
		}



			
		sbyt = mtpav_getreg( SREG );
				
	} while (sbyt & SIGS_BYTE);



	mtp_card->inmidistate = MTPAV_NO_CMD;

	if( rc == B_INVOKE_SCHEDULER ) //kprintf("\n");
				
	UNDEBUG
	return rc;
}

void queue_byte(uint8 byte) {
		NODEBUG
		mtpPort* port = mtp_card->ports[ translate_hwport_to_portindex( mtp_card->inmidiport )  ];

		if ( PORTREADOPEN( port ) ) {

			if( ++(port->lastbyte) == DATA_BUF_SIZE ) port->lastbyte = 0;
			
			port->databuf[ port->lastbyte ] = byte;
			kprintf("push %x %x\n", port->firstbyte, port->lastbyte );
			
			if( port->lastbyte+1 == port->firstbyte  ||  port->lastbyte +1 - DATA_BUF_SIZE == port->firstbyte ) // data buffer is full
				kprintf(" data buffer for port %d is full\n", mtp_card->inmidiport );	
							
			release_sem_etc( port->read_sem, 1, B_DO_NOT_RESCHEDULE );
		}			
		UNDEBUG
}



void mtpav_enable_interrupt() {
	NODEBUG
	uint8 control;

	kprintf("mtpav_enable_interrupt\n");
	
	// register interrupt handler with OS
	if ( B_ERROR == install_io_interrupt_handler ( MTPAV_IRQ, mtpav_interrupt, mtp_card, 0 ) )
		kprintf("install_io_interrupt_handler failed!\n");

	//set Interrupt Enable bit
	control = mtpav_getreg( CREG );
	control |= SIGC_INTEN;
	mtpav_putcreg( control );
	UNDEBUG
}

void mtpav_disable_interrupt() {
	NODEBUG
    uint8 control;

	kprintf( "mtpav_disable_interrupt\n" );
	    
    // clear Interrupt Enable bit
	control = mtpav_getreg( CREG );
	control &= (~SIGC_INTEN);
	mtpav_putcreg(control);
	
	// unregister interrupt handler from system
	if( B_ERROR == remove_io_interrupt_handler ( MTPAV_IRQ, mtpav_interrupt, mtp_card) )
		kprintf("remove_io_interrupt_handler failed!\n");

	UNDEBUG
}



uint8 mtpav_getreg(uint16 reg)
{
	NODEBUG
	uint8 rval = 0;

	if (reg == SREG) {
		rval = inb( (uint16)(MTPAV_IOBASE + SREG) );
		rval = (rval & SREG_MASK); //masking off unused bits of register
		//kprintf("get sreg: %2.2x\n", rval);

	} else if (reg == CREG) {
		rval = inb( (uint16)MTPAV_IOBASE + CREG );
		rval = (rval & CREG_MASK); //masking off  unused bits of register
		//kprintf("get creg: %2.2x\n", rval);

	}

	UNDEBUG	
	return rval;
}


void mtpav_putcreg( uint8 val ) {
	outb( MTPAV_IOBASE + CREG, val );
}


void mtpav_putdreg( uint8 val ) {
	outb( MTPAV_IOBASE + DREG, val );
}


void mtpav_send_byte(uint8 byte)
{
	NODEBUG
	uint8 tcbyt;
	uint8 clrwrite;
	uint8 setwrite;

	//kprintf("mtpav_send_byte %2.2x\n", byte );
	mtpav_wait_rfdhi();	// wait for "request-for-data" 

	// prepare control register values with write bit low/hi for strobe
	tcbyt = mtpav_getreg(CREG);
	clrwrite = tcbyt & (SIGC_WRITE ^ 0xff);
	setwrite = tcbyt | SIGC_WRITE;

	mtpav_putdreg( byte );
	mtpav_putcreg( clrwrite );	// write strobe
	mtpav_putcreg( setwrite );	// !write strobe

	UNDEBUG
}


uint8 inb( uint16 reg ) {
	return isa->read_io_8(reg);		
}

status_t outb( int reg, const uint8 val) {
	isa->write_io_8(reg, val);
	return B_OK;
}

int translate_hwport_to_portindex(int subdev)
{
	NODEBUG
	int pi = 0;
	
	if( subdev <= 0x08) pi = subdev; // single MTP's ports
   
    else if ( (subdev >= 0x09) && (subdev <= 0x10) ) pi = subdev + 8; // second MTP ports 
    else if ( subdev == 0x11 ) pi = 0x63; // ADAT 
	else pi = 0;		// all ports 

	UNDEBUG
	return pi;
}

int translate_devicename_to_portindex( const char* name ) {
	int port;
	
	if ( !strcmp( name, DEVICE_NAME_MIDI1 ) ) port = 1;
	else if ( !strcmp( name, DEVICE_NAME_MIDI2 ) ) port = 2;
	else if ( !strcmp( name, DEVICE_NAME_MIDI3 ) ) port = 3;
	else if ( !strcmp( name, DEVICE_NAME_MIDI4 ) ) port = 4;
	else if ( !strcmp( name, DEVICE_NAME_MIDI5 ) ) port = 5;
	else if ( !strcmp( name, DEVICE_NAME_MIDI6 ) ) port = 6;
	else if ( !strcmp( name, DEVICE_NAME_MIDI7 ) ) port = 7;
	else if ( !strcmp( name, DEVICE_NAME_MIDI8 ) ) port = 8;
	else port = 1;
	
	//kprintf("port index: %d\n", port);

	return port;
}